% This code is part of the article:
% How to tackle underdeterminacy in metabolic flux analysis? A tutorial and critical review
% copyright Ph. Bogaerts, A.Vande Wouwer
% Processes, 2021

clear all

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Toy example (vm1 = 3.5 and v2+v3 <= 5)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

N=[1 0 0 -1 0 0 0 -1 
    0 1 0 0 0 -1 0 0
    0 0 1 0 0 0 0 -1
    0 0 0 1 -1 0 0 0
    0 0 0 1 1 1 -1 0];
Aeq=[N ; 1 0 0 0 0 0 0 0];
beq=[0 0 0 0 0 3.5]';
A=[-eye(8) ; 0 1 1 0 0 0 0 0];
b=[0 0 0 0 0 0 0 0 5]';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FVA in the v space (vm1 = 3.5 and v2+v3 <= 5)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

size_v=size(A,2);
vmin=zeros(size_v,1);
vmax=zeros(size_v,1);
for i=1:size_v
    f=zeros(size_v,1);
    f(i)=1;
    [x,fval]=linprog(f,A,b,Aeq,beq);
    vmin(i,1)=fval;

    f=zeros(size_v,1);
    f(i)=-1;
    [x,fval]=linprog(f,A,b,Aeq,beq);
    vmax(i,1)=-fval;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Definition of the q space
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

v0=quadprog(eye(size_v),zeros(size_v,1),A,b,Aeq,beq);
Z=null(Aeq);
A2=A*Z;
b2=b-A*v0;

%%%%%%%%%%%%%%%%%%%%%
% FVA in the q space
%%%%%%%%%%%%%%%%%%%%%

size_q=size(A2,2);
qmin=zeros(size_q,1);
qmax=zeros(size_q,1);
for i=1:size_q
    f=zeros(size_q,1);
    f(i)=1;
    [x,fval]=linprog(f,A2,b2);
    qmin(i,1)=fval;

    f=zeros(size_q,1);
    f(i)=-1;
    [x,fval]=linprog(f,A2,b2);
    qmax(i,1)=-fval;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computation of samples and mean with rejection algorithm
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(0)
q_RA=rejection_algorithm(A2,b2,1e4);
q_RA_mean=mean(q_RA,2);

figure; plot(q_RA(1,:),q_RA(2,:),'.m'); hold; 
plot(q_RA_mean(1),q_RA_mean(2),'xk','LineWidth',2,'MarkerSize',15); xlabel('q_1'); ylabel('q_2')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computation of samples and mean with hit-and-run algorithm
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

q0=zeros(size_q,1);
q_HR=hit_and_run(A2,b2,q0,1e4);
q_HR_mean=mean(q_HR,2);

figure; plot(q_HR(1,:),q_HR(2,:),'.m'); hold; 
plot(q_HR_mean(1),q_HR_mean(2),'xk','LineWidth',2,'MarkerSize',15); xlabel('q_1'); ylabel('q_2')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computation of marginal distributions in the v space
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Marginal distributions from rejection algorithm samples
v=v0+Z*q_RA;
steps_v=10;
mpdf_v=marginal_distributions(v,vmin,vmax,steps_v);
figure;
for i=2:8
    subplot(2,4,i); histogram('BinEdges',linspace(vmin(i),vmax(i),steps_v+1),'BinCounts',mpdf_v(i,:),'DisplayStyle','bar','EdgeColor','blue');
    xlabel(strcat('v',num2str(i))); ylabel(strcat('mpdf(v',num2str(i),')'))
    axis([vmin(i) vmax(i) 0 max(mpdf_v(i,:))])
end
mpdf_v_1=mpdf_v;

% Marginal distributions from hit-and-run algorithm samples
v=v0+Z*q_HR;
steps_v=10;
mpdf_v=marginal_distributions(v,vmin,vmax,steps_v);
figure;
for i=2:8
    subplot(2,4,i); histogram('BinEdges',linspace(vmin(i),vmax(i),steps_v+1),'BinCounts',mpdf_v(i,:),'DisplayStyle','bar','EdgeColor','blue');
    xlabel(strcat('v',num2str(i))); ylabel(strcat('mpdf(v',num2str(i),')'))
    axis([vmin(i) vmax(i) 0 max(mpdf_v(i,:))])
end
mpdf_v_2=mpdf_v;

figure;
for i=2:8
    subplot(2,4,i); 
    h1=histogram('BinEdges',linspace(vmin(i),vmax(i),steps_v+1),'BinCounts',mpdf_v_1(i,:));
    h1.FaceColor='r';
    hold on
    h2=histogram('BinEdges',linspace(vmin(i),vmax(i),steps_v+1),'BinCounts',mpdf_v_2(i,:));
    h2.FaceColor='b';
    xlabel(strcat('v',num2str(i))); ylabel(strcat('mpdf(v',num2str(i),')'))
    axis([vmin(i) vmax(i) 0 max([mpdf_v_1(i,:) mpdf_v_2(i,:)])])
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FBA in the v space : maximizing v7
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

size_v=size(A,2);
f=zeros(size_v,1);
f(7)=-1;
[x,fval]=linprog(f,A,b,Aeq,beq);
v7_opt=-fval;
Aeq_FBA=[Aeq ; zeros(1,6) 1 0];
beq_FBA=[beq ; v7_opt];

vmin_FBA=zeros(size_v,1);
vmax_FBA=zeros(size_v,1);
for i=1:size_v
    f=zeros(size_v,1);
    f(i)=1;
    [x,fval]=linprog(f,A,b,Aeq_FBA,beq_FBA);
    vmin_FBA(i,1)=fval;

    f=zeros(size_v,1);
    f(i)=-1;
    [x,fval]=linprog(f,A,b,Aeq_FBA,beq_FBA);
    vmax_FBA(i,1)=-fval;
end

[vmin vmax vmin_FBA vmax_FBA]

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FBA in the v space : maximizing v8
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

size_v=size(A,2);
f=zeros(size_v,1);
f(8)=-1;
[x,fval]=linprog(f,A,b,Aeq,beq);
v8_opt=-fval;
Aeq_FBA=[Aeq ; zeros(1,7) 1];
beq_FBA=[beq ; v8_opt];

vmin_FBA=zeros(size_v,1);
vmax_FBA=zeros(size_v,1);
for i=1:size_v
    f=zeros(size_v,1);
    f(i)=1;
    [x,fval]=linprog(f,A,b,Aeq_FBA,beq_FBA);
    vmin_FBA(i,1)=fval;

    f=zeros(size_v,1);
    f(i)=-1;
    [x,fval]=linprog(f,A,b,Aeq_FBA,beq_FBA);
    vmax_FBA(i,1)=-fval;
end

[vmin vmax vmin_FBA vmax_FBA]

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FVA in the v space (v2 <= 5 , v5 <= 5 , v8 <= 5 and 0.95*10.5 <= v1 <= 1.05*10.5)
% which corresponds to a "noisy" measurement vm1 = 10.5 with smallest e_v = 5%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Aeq=[N];
beq=[0 0 0 0 0]';
A=[-eye(8) ; 0 1 0 0 0 0 0 0 ; 0 0 0 0 1 0 0 0 ; 0 0 0 0 0 0 0 1 ; 1 0 0 0 0 0 0 0 ; -1 0 0 0 0 0 0 0];
b=[0 0 0 0 0 0 0 0 5 5 5 1.05*10.5 -0.95*10.5]';

% REMARK : Using Aeq=[N ; 1 0 0 0 0 0 0 0] and beq=[0 0 0 0 0 10.5]' and
% withdrawing the last two inequalities accounting for 0.95*10.5 <= v1 <=
% 1.05*10.5 would lead to "No feasible solution found. 
% Linprog stopped because no point satisfies the constraints.".

size_v=size(A,2);
vmin=zeros(size_v,1);
vmax=zeros(size_v,1);
for i=1:size_v
    f=zeros(size_v,1);
    f(i)=1;
    [x,fval]=linprog(f,A,b,Aeq,beq);
    vmin(i,1)=fval;

    f=zeros(size_v,1);
    f(i)=-1;
    [x,fval]=linprog(f,A,b,Aeq,beq);
    vmax(i,1)=-fval;
end

[vmin vmax]
